# Copyright 2018 OpenStack Foundation.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""workbook_empty_namespace

Revision ID: 029
Revises: 028
Create Date: 2018-08-08 10:30:00.727769

"""

# revision identifiers, used by Alembic.
revision = '029'
down_revision = '028'

from alembic import op
import sqlalchemy as sa
from sqlalchemy.sql import table, column


def upgrade():
    wb_def = table('workbooks_v2', column('namespace'))
    session = sa.orm.Session(bind=op.get_bind())
    with session.begin(subtransactions=True):
        session.execute(
            wb_def.update().values(namespace='').where(
                wb_def.c.namespace == None))  # noqa

    session.commit()
