# Copyright 2019 - Nokia Corporation
#
#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.

from oslo_config import cfg


def is_validation_enabled(validate):
    validation_mode = cfg.CONF.api.validation_mode

    if validation_mode == 'mandatory':
        result = True
    elif validation_mode == 'disabled':
        result = False
    else:
        # validation_mode = 'enabled'
        result = validate

    return result
