# Copyright 2018 - Nokia, Inc.
#
#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.

from wsme import exc as wsme_exc

from mistral.tests.unit import base
from mistral.utils import rest_utils


class RestUtilsTest(base.BaseTest):
    def test_validate_fields(self):
        rest_utils.validate_fields(["a", "b"], ["a", "b", "c"])

        e = self.assertRaises(wsme_exc.ClientSideError,
                              rest_utils.validate_fields, ["d"], ["a"])
        self.assertIn("[d]", str(e))
